# -*- coding: utf-8 -*-
# Copyright (C) 2020 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.modules.quorum"""

import os
import re
import logging
import time
import uuid

from datetime import datetime

from copy import deepcopy

import requests
from six import iteritems

from botocore.exceptions import ClientError, WaiterError

from dwho.classes.modules import MODULES
from dwho.classes.notifiers import DWhoPushNotifications

from httpdis.ext.httpdis_json import HttpReqErrJson

from sonicprobe import helpers
from sonicprobe.libs import network, urisup, xys
from sonicprobe.libs.moresynchro import RWLock

from ..classes.common import (DATE_FORMAT,
                              DB_REPLICATION_TYPE_DEFAULT,
                              DB_REPLICATION_TYPE_OPTS,
                              STATUS_FAILURE,
                              STATUS_PROCESSING,
                              STATUS_SUCCESS,
                              STATUS_STOPPED)

from ..classes.abstract.modules import XtrargonModuleBase
from ..classes.abstract.db import XtrargonDBBase
from ..classes.ec2 import XtrargonEc2Helper

LOG                         = logging.getLogger('xtrargon_ng.modules.quorum')

DB_NODE_LETTERS             = {'a': 'b',
                               'b': 'a'}

DEFAULT_CNF_FILE            = "quorum.cnf"

DEFAULT_BACKUPS_STATUS_URI  = "http://:0/backups/status"
DEFAULT_FORBID_VIP_URI      = "http://:0/server/forbid-vip"
DEFAULT_MAINT_URI           = "http://:0/server/maint"
DEFAULT_REQUEST_TIMEOUT     = 15

CLUSTER_VAR_PATH            = "/var/lib/xtrargon-ng/quorum-cluster"

CLUSTER_FILE_PATTERN        = "cluster_%s"

MATCH_DB_NODE               = re.compile(r'^([a-z0-9][a-z0-9\-\.]+[a-z0-9]-[0-9]+)([a-z])(\.[a-z0-9\-\.]+[a-z0-9]|)$').match

SNAPSHOT_MAX_RETENTION_DAYS = 30

SNAPSHOT_TAG_DB_BKP         = 'ct-db-backup'
SNAPSHOT_TAG_DB_CLUSTER_BKP = 'ct-db-cluster-backup'


xys.add_callback('quorum.host', network.valid_domain)

class QuorumDB(XtrargonDBBase): # pylint: disable-msg=too-few-public-methods
    pass

class QuorumModule(XtrargonModuleBase):
    MODULE_NAME = 'quorum'

    LOCK = RWLock()

    @staticmethod
    def _load_ec2helper(config = None):
        return XtrargonEc2Helper(config).load_credentials(section = 'ec2')

    # pylint: disable-msg=attribute-defined-outside-init
    def safe_init(self, options):
        self.lock_timeout = self.config['general']['lock_timeout']
        self.request_timeout = DEFAULT_REQUEST_TIMEOUT

        backups_status_uri = DEFAULT_BACKUPS_STATUS_URI
        forbid_vip_uri = DEFAULT_FORBID_VIP_URI
        maint_uri = DEFAULT_MAINT_URI

        self.snapshot_max_retention = SNAPSHOT_MAX_RETENTION_DAYS

        self._clusters = {}

        self.reset_state()

        if not isinstance(self.config.get('quorum'), dict):
            self.ec2helper = self._load_ec2helper()
        else:
            self.ec2helper = self._load_ec2helper(self.config['quorum'])

            if self.config['quorum'].get('backups_status_uri'):
                backups_status_uri = self.config['quorum']['backups_status_uri']

            if self.config['quorum'].get('forbid_vip_uri'):
                forbid_vip_uri = self.config['quorum']['forbid_vip_uri']

            if self.config['quorum'].get('maint_uri'):
                maint_uri = self.config['quorum']['maint_uri']

            if 'request_timeout' in self.config['quorum']:
                self.request_timeout = float(self.config['quorum']['request_timeout'])

            if self.config['quorum'].get('clusters'):
                self._clusters = self.config['quorum']['clusters']

            self.snapshot_max_retention = int(self.config['quorum'].get('snapshot_max_retention_days') or 0) \
                                          or SNAPSHOT_MAX_RETENTION_DAYS

        self.backups_status_uri = list(urisup.uri_help_split(backups_status_uri))
        self.forbid_vip_uri = list(urisup.uri_help_split(forbid_vip_uri))
        self.maint_uri = list(urisup.uri_help_split(maint_uri))

        if not os.path.isdir(CLUSTER_VAR_PATH):
            helpers.make_dirs(CLUSTER_VAR_PATH)

        self.cluster_freeze_file = os.path.join(CLUSTER_VAR_PATH,
                                                "%s.freeze" % CLUSTER_FILE_PATTERN)

        self.notifiers = None
        if self.config['general'].get('notifiers_path'):
            self.notifiers = DWhoPushNotifications(self.server_id,
                                                   self.config['general']['notifiers_path'])

    def get_state(self, status = None, end = False):
        if 'status' not in self._state:
            self.reset_state()

        self._state['errors'] = self._errors

        if status:
            self._state['status'] = status

        started_at = self._state.get('started_at')
        if started_at:
            started_at = datetime.strptime(started_at, DATE_FORMAT)
            duration   = (datetime.now() - started_at).total_seconds()
            self._state['duration'] = duration

        if not end or self._state['ended_at']:
            return self._state

        self._state['ended_at'] = time.strftime(DATE_FORMAT)

        if self._state['status'] == STATUS_STOPPED:
            self._state['errors'].append('stopped before completing')
        elif not self.has_errors():
            self._state['status'] = STATUS_SUCCESS
        else:
            self._state['status'] = STATUS_FAILURE

        return self._state

    @staticmethod
    def _get_neighbor_node(host):
        node = host.lower()

        m = MATCH_DB_NODE(host)
        if not m:
            raise HttpReqErrJson(415, "invalid node host argument: %r" % node)

        letter = m.group(2)
        if letter not in DB_NODE_LETTERS:
            raise HttpReqErrJson(415, "invalid node host letter: %r" % node)

        host_splitted    = list(m.groups())
        host_splitted[1] = DB_NODE_LETTERS[letter]

        return ''.join(host_splitted)

    def _get_cluster_info(self, cluster):
        cluster = cluster.lower()
        info = deepcopy(self._clusters.get(cluster))

        if not info:
            return None

        info['host'] = cluster

        if info.get('replication_type') is None:
            info['replication_type'] = DB_REPLICATION_TYPE_DEFAULT
        elif info.get('replication_type') in DB_REPLICATION_TYPE_OPTS:
            info['replication_type'] = info.get('replication_type')
        else:
            info['replication_type'] = bool(info.get('replication_type'))

        return info

    def _build_uri(self, uri, host = None):
        uri = list(uri)
        if uri[1]:
            uri[1] = list(uri[1])
        else:
            uri[1] = [None, None, None, None]

        if host and not uri[1][2]:
            uri[1][2] = host

        if uri[1][3] == '0':
            uri[1][3] = self.config['general'].get('listen_port')

        return urisup.uri_help_unsplit(uri)

    def _get_routes_by_instance(self, cluster, client = None):
        r = {}

        if not client:
            client = self.ec2helper.boto3_client()

        instances = self.ec2helper.instance_info(
            client.describe_instances(
                Filters = [{'Name': 'tag:PrivateDnsName',
                            'Values': cluster['instances']}]))

        if not instances:
            LOG.error("unable to find any instance")
            return None

        for name, value in iteritems(instances):
            r[name] = {'id': value['id'],
                       'routes': {}}

            rtbs = client.describe_route_tables(
                Filters = [{'Name': 'route.instance-id',
                            'Values': [value['id']]}],
                RouteTableIds = cluster['routes'])

            if not rtbs or not rtbs.get('RouteTables'):
                LOG.debug("unable to find a route for instance: %r", name)
                continue

            for rtb in rtbs['RouteTables']:
                for route in rtb['Routes']:
                    if route \
                       and route.get('DestinationCidrBlock') \
                       and route['DestinationCidrBlock'].startswith("%s/" % cluster['vip']):
                        if rtb['RouteTableId'] not in r[name]['routes']:
                            r[name]['routes'][rtb['RouteTableId']] = []
                        r[name]['routes'][rtb['RouteTableId']].append(route)

        return r


    def _prepare_replace_routes(self, cluster, client = None, replica = True):
        r = {'instance_id': None,
             'instance_name': None,
             'replication_type': None,
             'routed_as': None,
             'routes': []}

        routes_instance = self._get_routes_by_instance(cluster, client)
        if not routes_instance:
            return None

        for name, value in iteritems(routes_instance):
            if not value['routes'] and replica:
                r['instance_id'] = value['id']
                r['instance_name'] = name
                r['replication_type'] = cluster['replication_type']
                r['routed_as'] = 'replica'
                continue

            if value['routes'] and not replica:
                r['instance_id'] = value['id']
                r['instance_name'] = name
                r['replication_type'] = cluster['replication_type']
                r['routed_as'] = 'primary'

            if not value['routes']:
                continue

            for route, info in iteritems(value['routes']):
                for x in info:
                    r['routes'].append({'DestinationCidrBlock': x['DestinationCidrBlock'],
                                        'RouteTableId': route})

        if not r['instance_id']:
            return False

        return r

    def _touch_freeze_cluster(self, host):
        helpers.touch(self.cluster_freeze_file % host)

    def _exists_freeze_cluster(self, host):
        xfile = self.cluster_freeze_file % host
        return os.path.exists(xfile)

    def _unlink_freeze_cluster(self, host):
        xfile = self.cluster_freeze_file % host
        if os.path.exists(xfile):
            os.unlink(xfile)

    def _is_maint_server(self, host):
        req = None

        try:
            LOG.info("get maintenance status on %r", host)

            req = requests.get(self._build_uri(self.maint_uri, host),
                               timeout = self.request_timeout)
            if not req.text:
                return None

            return req.json()
        except Exception as e:
            self.add_errors(repr(e))
            LOG.exception(e)
        finally:
            if req:
                req.close()

        return None

    def _set_maint_server(self, host):
        req = None

        try:
            LOG.info("set maintenance on %r", host)

            req = requests.post(self._build_uri(self.maint_uri, host),
                                timeout = self.request_timeout)
            if not req.text:
                return None

            return req.json()
        except Exception as e:
            self.add_errors(repr(e))
            LOG.exception(e)
        finally:
            if req:
                req.close()

        return None

    def _unset_maint_server(self, host):
        req = None

        try:
            LOG.info("unset maintenance on %r", host)

            req = requests.delete(self._build_uri(self.maint_uri, host),
                                  timeout = self.request_timeout)
            if not req.text:
                return None

            return req.json()
        except Exception as e:
            self.add_errors(repr(e))
            LOG.exception(e)
        finally:
            if req:
                req.close()

        return None

    def _start_instances(self, client, cluster, data, dry_run = False):
        try:
            client.start_instances(InstanceIds=[data['instance_id']],
                                  DryRun = dry_run)
        except ClientError as e:
            if 'DryRunOperation' not in str(e):
                self.add_errors(repr(e))
                raise

        waiter = client.get_waiter('instance_running')

        try:
            waiter.wait(InstanceIds=[data['instance_id']],
                        DryRun = dry_run)
        except WaiterError as e:
            if 'DryRunOperation' not in str(e):
                self.add_errors(repr(e))
                LOG.error(e)
                raise HttpReqErrJson(500, "unable to start instance %r for cluster: %r"
                                          % (data['instance_name'], cluster['host']))

    def _stop_instances(self, client, cluster, data, dry_run = False):
        try:
            client.stop_instances(InstanceIds=[data['instance_id']],
                                  DryRun = dry_run)
        except ClientError as e:
            LOG.error("%r", e)
            if 'DryRunOperation' not in str(e):
                self.add_errors(repr(e))
                raise

        waiter = client.get_waiter('instance_stopped')

        try:
            waiter.wait(InstanceIds=[data['instance_id']],
                        DryRun = dry_run)
        except WaiterError as e:
            if 'DryRunOperation' not in str(e):
                self.add_errors(repr(e))
                LOG.error(e)
                raise HttpReqErrJson(500, "unable to stop instance %r for cluster: %r"
                                          % (data['instance_name'], cluster['host']))

    def _create_snapshots(self, client, cluster, data, dry_run = False):
        rs = None
        snapshots = []

        try:
            rs = client.create_snapshots(
                Description = ("xtrargon instance snapshots %r for cluster: %r"
                               % (data['instance_name'], cluster['host'])),
                InstanceSpecification = {'InstanceId': data['instance_id'],
                                         'ExcludeBootVolume': False},
                TagSpecifications = [{'ResourceType': 'snapshot',
                                      'Tags': [{'Key': 'Name',
                                                'Value': data['instance_name']},
                                               {'Key': SNAPSHOT_TAG_DB_BKP,
                                                'Value': data['instance_id']},
                                               {'Key': SNAPSHOT_TAG_DB_CLUSTER_BKP,
                                                'Value': cluster['host']},
                                               {'Key': 'Backups',
                                                'Value': 'xtrargon'}]}],
                DryRun = dry_run)
        except ClientError as e:
            if 'DryRunOperation' not in str(e):
                self.add_errors(repr(e))
                raise
        finally:
            try:
                time.sleep(5)
                self._start_instances(client, cluster, data, dry_run = dry_run)
            except Exception as e:
                LOG.exception(e)

        if not rs or not rs.get('Snapshots'):
            return None

        for snapshot in rs['Snapshots']:
            snapshots.append(snapshot['SnapshotId'])

        waiter = client.get_waiter('snapshot_completed')

        try:
            waiter.wait(SnapshotIds = snapshots,
                        WaiterConfig = {'Delay': 30,
                                        'MaxAttempts': 90},
                        DryRun = dry_run)
        except WaiterError as e:
            if 'DryRunOperation' not in str(e):
                self.add_errors(repr(e))
                LOG.error(e)
                raise HttpReqErrJson(500, "unable to snapshots instance %r for cluster: %r"
                                          % (data['instance_name'], cluster['host']))

        return snapshots

    @staticmethod
    def _delete_old_snapshots(client, cluster, data, dry_run = False):
        rs = None

        try:
            rs = client.describe_snapshots(
                    Filters = [{'Name': "tag:%s" % SNAPSHOT_TAG_DB_CLUSTER_BKP,
                                'Values': [cluster['host']]}],
                DryRun = dry_run)
        except ClientError as e:
            if 'DryRunOperation' not in str(e):
                raise

        if not rs or not rs.get('Snapshots'):
            return

        for snapshot in rs['Snapshots']:
            d = datetime.now().date() - snapshot['StartTime'].date()
            if d.days <= SNAPSHOT_MAX_RETENTION_DAYS:
                continue

            try:
                client.delete_snapshot(SnapshotId = snapshot['SnapshotId'],
                                       DryRun = dry_run)
            except ClientError as e:
                if 'InvalidSnapshot.InUse' in str(e):
                    LOG.warning("unable to delete this snapshot %r instance %r for cluster: %r",
                                snapshot['SnapshotId'],
                                data['instance_name'],
                                cluster['host'])
                    continue

                if 'DryRunOperation' not in str(e):
                    LOG.exception(e)
                    continue


    BACKUPS_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def backups_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.BACKUPS_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        host = self._get_neighbor_node(params['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        req = None

        try:
            LOG.info("get backups status on %r", host)

            req = requests.get(self._build_uri(self.backups_status_uri, host),
                               timeout = self.request_timeout)
            if not req.text:
                return None

            try:
                return req.json()
            except ValueError:
                req.raise_for_status()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if req:
                req.close()
            self.LOCK.release()

        return None


    FORBID_VIP_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def forbid_vip_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.FORBID_VIP_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        host = self._get_neighbor_node(params['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        req = None

        try:
            LOG.info("get forbid vip status on %r", host)

            req = requests.get(self._build_uri(self.forbid_vip_uri, host),
                               timeout = self.request_timeout)
            if not req.text:
                return None

            try:
                return req.json()
            except ValueError:
                req.raise_for_status()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if req:
                req.close()
            self.LOCK.release()

        return None


    REPLICA_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def replica_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.REPLICA_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        host = self._get_neighbor_node(params['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        qdb = None

        try:
            LOG.info("get replica status on %r", host)
            qdb = QuorumDB()
            qdb.init_db(module = self, default_cnf_file = DEFAULT_CNF_FILE, host = host)
            qdb.create_connection()

            return qdb.get_slave_status()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if qdb:
                qdb.close_connection()
            self.LOCK.release()


    CHECK_CLUSTER_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)
    def check_cluster_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.CHECK_CLUSTER_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        cluster = self._get_cluster_info(params['host'])
        if not cluster:
            raise HttpReqErrJson(415, "invalid cluster name: %r" % params['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        qdb = None

        try:
            client = self.ec2helper.boto3_client()

            r = self._prepare_replace_routes(cluster, client, replica = False)
            if r is None:
                raise HttpReqErrJson(404, "unable to find routes for cluster: %r" % cluster['host'])

            if r['replication_type'] == 'primary-only':
                raise HttpReqErrJson(501,
                                     "routed server can not be a replica on %r. (replication: %r)"
                                     % (r['instance_name'], r['replication_type']))

            if r['replication_type'] == 'replica-only':
                raise HttpReqErrJson(502,
                                     "routed server can not be a primary on %r. (replication: %r)"
                                     % (r['instance_name'], r['replication_type']))

            if r['replication_type'] != 'primary-replica':
                raise HttpReqErrJson(503,
                                     "routed server can not be a primary on %r. (replication: %r)"
                                     % (r['instance_name'], r['replication_type']))

            LOG.info("get primary status on %r. (replication: %r)",
                     r['instance_name'],
                     r['replication_type'])
            qdb = QuorumDB()
            qdb.init_db(module = self, default_cnf_file = DEFAULT_CNF_FILE, host = r['instance_name'])
            qdb.create_connection()

            if qdb.check_is_slave():
                raise HttpReqErrJson(500,
                                     "routed server is replica on %r. (replication: %r)"
                                     % (r['instance_name'], r['replication_type']))

            if not qdb.check_is_master():
                raise HttpReqErrJson(412,
                                     "routed server is not primary on %r. (replication: %r)"
                                     % (r['instance_name'], r['replication_type']))

            return r
        except HttpReqErrJson:
            raise
        except Exception as e:
            raise HttpReqErrJson(503, repr(e))
        finally:
            if self.ec2helper:
                self.ec2helper.terminate()
            if qdb:
                qdb.close_connection()
            self.LOCK.release()


    FAILOVER_REPLICA_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    dry_run?: !~~isBool
    freeze?: !~~isBool
    """)

    def failover_replica(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.FAILOVER_REPLICA_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        cluster = self._get_cluster_info(params['host'])
        if not cluster:
            raise HttpReqErrJson(415, "invalid cluster name: %r" % params['host'])

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        qdb = None

        try:
            client = self.ec2helper.boto3_client()

            r = self._prepare_replace_routes(cluster, client)
            if r is None:
                raise HttpReqErrJson(404, "unable to find routes to failover for cluster: %r" % cluster['host'])

            if r is False:
                raise HttpReqErrJson(500, "unable to find instance to failover for cluster: %r" % cluster['host'])

            LOG.info("get replica status on %r", r['instance_name'])
            qdb = QuorumDB()
            qdb.init_db(module = self, default_cnf_file = DEFAULT_CNF_FILE, host = r['instance_name'])
            qdb.create_connection()

            if r['replication_type'] == 'primary-only':
                raise HttpReqErrJson(501,
                                     "server can not be a replica on %r. (replication: %r)"
                                     % (r['instance_name'], r['replication_type']))

            if r['replication_type'] == 'replica-only':
                raise HttpReqErrJson(502,
                                     "server can not be a primary on %r. (replication: %r)"
                                     % (r['instance_name'], r['replication_type']))

            if r['replication_type'] != 'primary-replica':
                raise HttpReqErrJson(503,
                                     "replication type is not primary-replica on %r. (replication: %r)"
                                     % (r['instance_name'], r['replication_type']))

            if qdb.check_is_slave() != 'Yes':
                raise HttpReqErrJson(504, "SQL replica thread isn't running on %r" % r['instance_name'])

            #if not qdb.replica_io_running():
            #    raise HttpReqErrJson(505, "SQL replica IO isn't running on %r" % r['instance_name'])

            if qdb.check_is_master():
                raise HttpReqErrJson(506, "SQL replica is also primary on %r" % r['instance_name'])

            if helpers.boolize(params.get('dry_run')):
                return r

            for route in r['routes']:
                route['InstanceId'] = r['instance_id']
                client.replace_route(**route)

            self._touch_freeze_cluster(cluster['host'])

            return r
        except HttpReqErrJson:
            raise
        except Exception as e:
            raise HttpReqErrJson(503, repr(e))
        finally:
            if self.ec2helper:
                self.ec2helper.terminate()
            if qdb:
                qdb.close_connection()
            self.LOCK.release()


    FREEZE_CLUSTER_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def freeze_cluster(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.FREEZE_CLUSTER_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        cluster = self._get_cluster_info(params['host'])
        if not cluster:
            raise HttpReqErrJson(415, "invalid cluster name: %r" % params['host'])

        if request.get_method() == 'GET':
            if not self.LOCK.acquire_read(self.lock_timeout):
                raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)
        else:
            if not self.LOCK.acquire_write(self.lock_timeout):
                raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        r = True

        try:
            if request.get_method() == 'DELETE':
                LOG.info("unlink freeze file for cluster %r", cluster['host'])
                self._unlink_freeze_cluster(cluster['host'])
            elif request.get_method() == 'GET':
                LOG.info("check freeze file exists for cluster %r", cluster['host'])
                r = self._exists_freeze_cluster(cluster['host'])
            elif request.get_method() == 'POST':
                LOG.info("write freeze file for cluster %r", cluster['host'])
                self._touch_freeze_cluster(cluster['host'])
            else:
                raise HttpReqErrJson(400, "invalid request method")

            return r
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    ROUTE_TABLES_FORCE_REPLICA_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    dry_run?: !~~isBool
    freeze?: !~~isBool
    """)
    def route_tables_force_replica(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.ROUTE_TABLES_FORCE_REPLICA_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        cluster = self._get_cluster_info(params['host'])
        if not cluster:
            raise HttpReqErrJson(415, "invalid cluster name: %r" % params['host'])

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            client = self.ec2helper.boto3_client()

            r = self._prepare_replace_routes(cluster, client)
            if r is None:
                raise HttpReqErrJson(404, "unable to find routes to failover for cluster: %r" % cluster['host'])

            if r is False:
                raise HttpReqErrJson(500, "unable to find instance to failover for cluster: %r" % cluster['host'])

            if helpers.boolize(params.get('dry_run')):
                return r['routes']

            for route in r['routes']:
                route['InstanceId'] = r['instance_id']
                client.replace_route(**route)

            if helpers.boolize(params.get('freeze')):
                self._touch_freeze_cluster(cluster['host'])
            else:
                self._unlink_freeze_cluster(cluster['host'])

            return r['routes']
        except HttpReqErrJson:
            raise
        except Exception as e:
            raise HttpReqErrJson(503, repr(e))
        finally:
            if self.ec2helper:
                self.ec2helper.terminate()
            self.LOCK.release()


    ROUTE_TABLES_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def route_tables_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.ROUTE_TABLES_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        cluster = self._get_cluster_info(params['host'])
        if not cluster:
            raise HttpReqErrJson(415, "invalid cluster name: %r" % params['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            r = self._get_routes_by_instance(cluster)
            if not r:
                raise HttpReqErrJson(404, "unable to find routes for cluster: %r" % cluster['host'])

            return r
        except HttpReqErrJson:
            raise
        except Exception as e:
            raise HttpReqErrJson(503, repr(e))
        finally:
            if self.ec2helper:
                self.ec2helper.terminate()
            self.LOCK.release()


    SNAPSHOT_CLUSTER_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    dry_run?: !~~isBool
    """)

    def snapshot_cluster(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.FAILOVER_REPLICA_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        cluster = self._get_cluster_info(params['host'])
        if not cluster:
            raise HttpReqErrJson(415, "invalid cluster name: %r" % params['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        dry_run = helpers.boolize(params.get('dry_run'))
        qdb = None
        is_maint = False

        self.set_state_value('started_at', time.strftime(DATE_FORMAT))
        self.set_state_value('uuid', "%s" % uuid.uuid4())
        self.set_state_value('status', STATUS_PROCESSING)
        self.set_state_value('server_id', cluster['host'])

        try:
            LOG.info("starting snapshot for cluster: %r", cluster['host'])

            if not dry_run and self.has_notifications():
                self.notifiers(xvars = self.get_state(),
                               tags = ('start-snapshot',))

            client = self.ec2helper.boto3_client()

            r = self._prepare_replace_routes(cluster, client)
            if r is None:
                raise HttpReqErrJson(404, "unable to find routes to snapshot for cluster: %r" % cluster['host'])

            if r is False:
                raise HttpReqErrJson(500, "unable to find instance to snapshot for cluster: %r" % cluster['host'])

            LOG.info("get replica status on %r", r['instance_name'])
            qdb = QuorumDB()
            qdb.init_db(module = self, default_cnf_file = DEFAULT_CNF_FILE, host = r['instance_name'])
            qdb.create_connection()

            if r['replication_type'] == 'primary-only':
                raise HttpReqErrJson(501,
                                     "server can not be a replica on %r. (replication: %r)"
                                     % (r['instance_name'], r['replication_type']))

            if r['replication_type'] == 'replica-only':
                raise HttpReqErrJson(502,
                                     "server can not be a primary on %r. (replication: %r)"
                                     % (r['instance_name'], r['replication_type']))

            if r['replication_type'] != 'primary-replica':
                raise HttpReqErrJson(503,
                                     "replication type is not primary-replica on %r. (replication: %r)"
                                     % (r['instance_name'], r['replication_type']))

            if qdb.check_is_slave() != 'Yes':
                raise HttpReqErrJson(504, "SQL replica thread isn't running on %r" % r['instance_name'])

            if not qdb.replica_io_running():
                raise HttpReqErrJson(505, "SQL replica IO isn't running on %r" % r['instance_name'])

            if qdb.check_is_master():
                raise HttpReqErrJson(506, "SQL replica is also primary on %r" % r['instance_name'])

            if not self._is_maint_server(r['instance_name']):
                is_maint = self._set_maint_server(r['instance_name'])
                time.sleep(30)

            self._stop_instances(client, cluster, r, dry_run = dry_run)

            self._create_snapshots(client, cluster, r, dry_run = dry_run)

            return r
        except HttpReqErrJson as e:
            self.add_errors(repr(e))
            raise
        except Exception as e:
            self.add_errors(repr(e))
            raise HttpReqErrJson(503, repr(e))
        finally:
            if is_maint:
                self._unset_maint_server(r['instance_name'])
            if self.ec2helper:
                self.ec2helper.terminate()
            if qdb:
                qdb.close_connection()
            self.LOCK.release()

            state = self.get_state(end = True)

            if not dry_run and self.has_notifications():
                self.notifiers(xvars = state,
                               tags = ('stop-snapshot',))


    VIP_CONNECTION_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def vip_connection_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.VIP_CONNECTION_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        cluster = self._get_cluster_info(params['host'])
        if not cluster:
            raise HttpReqErrJson(415, "invalid cluster name: %r" % params['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        qdb = None

        try:
            LOG.info("get VIP connection status on cluster %r", cluster['host'])
            qdb = QuorumDB()
            qdb.init_db(module = self, default_cnf_file = DEFAULT_CNF_FILE, host = cluster['vip'])
            qdb.create_connection()

            return qdb.test_connection()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if qdb:
                qdb.close_connection()
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(QuorumModule())
    _start()
